import logging

from .distributions import Normal
from .utils.optimization import relative_error, optimize_quartile

_log = logging.getLogger("preliz")


def quartile(
    distribution=None,
    q1=-1,
    q2=0,
    q3=1,
    plot=True,
    plot_kwargs=None,
    ax=None,
):
    """
    Find the maximum entropy distribution with `mass` in the interval defined by the `lower` and
    `upper` end-points.

    Parameters
    ----------
    distribution : PreliZ distribution
        Instance of a PreliZ distribution
    q1 : float
        First quartile, i.e 0.25 of the mass is below this point.
    q2 : float
        Second quartile, i.e 0.50 of the mass is below this point. This is also know
        as the median.
    q3 : float
        Third quartile, i.e 0.75 of the mass is below this point.
    plot : bool
        Whether to plot the distribution, and lower and upper bounds. Defaults to True.
    plot_kwargs : dict
        Dictionary passed to the method ``plot_pdf()`` of ``distribution``.
    ax : matplotlib axes

    Returns
    -------

    axes: matplotlib axes
    """
    if plot_kwargs is None:
        plot_kwargs = {}

    if not q1 < q2 < q3:
        raise ValueError("The order of the quartiles should be q1 < q2 < q3")

    if distribution is None:
        distribution = Normal()

    distribution._check_endpoints(q1, q3)

    # Heuristic to provide an initial guess for the optimization step
    # We obtain those guesses by first approximating the mean and standard deviation
    # from the quartiles and then use those values for moment matching
    distribution._fit_moments(mean=q2, sigma=(q3 - q1) * 1.5)  # pylint:disable=protected-access

    opt = optimize_quartile(distribution, (q1, q2, q3))

    r_error, computed_mass = relative_error(distribution, q1, q3, 0.5)

    if r_error > 0.01:
        _log.info(
            "The requested mass in the interval (q1=%.2g - q3=%.2g) is 0.5, "
            "but the computed one is %.2g",
            q1,
            q3,
            computed_mass,
        )

    if plot:
        ax = distribution.plot_pdf(**plot_kwargs)
        if plot_kwargs.get("pointinterval"):
            cid = -4
        else:
            cid = -1
        ax.plot([q1, q2, q3], [0, 0, 0], "o", color=ax.get_lines()[cid].get_c(), alpha=0.5)
    return ax, opt



"""
Parent classes for all families.
"""
# pylint: disable=no-member
from collections import namedtuple

import numpy as np

from ..utils.plot_utils import plot_pdfpmf, plot_cdf, plot_ppf
from ..utils.utils import hdi_from_pdf


class Distribution:
    """
    Base class for distributions.

    Not intended for direct instantiation.
    """

    def __init__(self):
        self.rv_frozen = None
        self.is_frozen = False
        self.opt = None

    def __repr__(self):
        if self.is_frozen:
            bolded_name = "\033[1m" + self.name.capitalize() + "\033[0m"

            # temporary patch until we migrate all distributions to use
            # self.params_report and self.params
            try:
                params_value = self.params_report
            except AttributeError:
                params_value = self.params
            description = "".join(
                f"{n}={v:.2f}," for n, v in zip(self.param_names, params_value)
            ).strip(",")
            return f"{bolded_name}({description})"
        else:
            return self.name

    def _update_rv_frozen(self):
        """Update the rv_frozen object"""

        frozen = self._get_frozen()
        if frozen is not None:
            self.is_frozen = True
            self.rv_frozen = frozen
            self.rv_frozen.name = self.name

    def summary(self, fmt=".2f", mass=0.94):
        """
        Namedtuple with the mean, median, standard deviation, and lower and upper bounds
        of the equal-tailed interval.

        Parameters
        ----------
        fmt : str
            fmt used to represent results using f-string fmt for floats. Default to ".2f"
            i.e. 2 digits after the decimal point.
        mass: float
            Probability mass for the equal-tailed interval. Defaults to 0.94
        """
        if self.is_frozen:
            attr = namedtuple(self.name.capitalize(), ["mean", "median", "std", "lower", "upper"])
            mean = float(f"{self.rv_frozen.mean():{fmt}}")
            median = float(f"{self.rv_frozen.median():{fmt}}")
            std = float(f"{self.rv_frozen.std():{fmt}}")
            eti = self.rv_frozen.interval(mass)
            lower_tail = float(f"{eti[0]:{fmt}}")
            upper_tail = float(f"{eti[1]:{fmt}}")
            return attr(mean, median, std, lower_tail, upper_tail)
        else:
            return None

    def rvs(self, size=1, random_state=None):
        """Random sample

        Parameters
        ----------
        size : int or tuple of ints, optional
            Defining number of random variates. Defaults to 1.
        random_state : {None, int, numpy.random.Generator, numpy.random.RandomState}
            Defaults to None
        """
        return self.rv_frozen.rvs(size=size, random_state=random_state)

    def eti(self, fmt=".2f", mass=0.94):
        """Equal-tailed interval containing `mass`.

        Parameters
        ----------
        fmt : str
            fmt used to represent results using f-string fmt for floats. Default to ".2f"
            i.e. 2 digits after the decimal point.
        mass: float
            Probability mass in the interval. Defaults to 0.94
        """
        eti = self.rv_frozen.interval(mass)
        lower_tail = float(f"{eti[0]:{fmt}}")
        upper_tail = float(f"{eti[1]:{fmt}}")
        return (lower_tail, upper_tail)

    def hdi(self, fmt=".2f", mass=0.94):
        """Highest density interval containing `mass`.

        Parameters
        ----------
        fmt : str
            fmt used to represent results using f-string fmt for floats. Default to ".2f"
            i.e. 2 digits after the decimal point.
        mass: float
            Probability mass in the interval. Defaults to 0.94
        """
        hdi = hdi_from_pdf(self, mass)
        lower_tail = float(f"{hdi[0]:{fmt}}")
        upper_tail = float(f"{hdi[1]:{fmt}}")
        return (lower_tail, upper_tail)

    def _check_endpoints(self, lower, upper, raise_error=True):
        """
        Evaluate if the lower and upper values are in the support of the distribution

        Parameters
        ----------
        lower : int or float
            lower endpoint
        upper : int or float
            upper endpoint
        raise_error : bool
            If True (default) it will raise ValueErrors, otherwise it will return True
            if the lower and upper endpoint are in the support of the distribution or
            False otherwise.
        """
        s_l, s_u = self.support

        if raise_error:
            domain_error = (
                f"The provided endpoints are outside the domain of the {self.name} distribution"
            )

            if np.isfinite(s_l):
                if lower < s_l:
                    raise ValueError(domain_error)

            if np.isfinite(s_u):
                if upper > s_u:
                    raise ValueError(domain_error)
            if np.isfinite(s_l) and np.isfinite(s_u):
                if lower == s_l and upper == s_u:
                    raise ValueError(
                        "Given the provided endpoints, mass will be always 1. "
                        "Please provide other values"
                    )
            return None
        else:
            return lower >= s_l and upper <= s_u

    def _finite_endpoints(self, support):
        """
        Return finite endpoints even for unbounded distributions

        Parameters
        ----------
        support : str or tuple
            Available string options are "full" or "restricted".
        """
        if isinstance(support, tuple):
            lower_ep, upper_ep = support
        else:
            lower_ep, upper_ep = self.support

            if not np.isfinite(lower_ep) or support == "restricted":
                lower_ep = self.rv_frozen.ppf(0.0001)
            if not np.isfinite(upper_ep) or support == "restricted":
                upper_ep = self.rv_frozen.ppf(0.9999)

        return lower_ep, upper_ep

    def plot_pdf(
        self,
        moments=None,
        pointinterval=False,
        quantiles=None,
        support="full",
        legend="legend",
        figsize=None,
        ax=None,
    ):
        """
        Plot the  pdf (continuous) or pmf (discrete).

        Parameters
        ----------
        moments : str
            Compute moments. Use any combination of the strings ``m``, ``d``, ``v``, ``s`` or ``k``
            for the mean (μ), standard deviation (σ), variance (σ²), skew (γ) or kurtosis (κ)
            respectively. Other strings will be ignored. Defaults to None.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        quantiles : list
            Values of the five quantiles to use when ``pointinterval=True`` if None (default)
            the values ``[0.05, 0.25, 0.5, 0.75, 0.95]`` will be used. The number of elements
            should be 5, 3, 1 or 0 (in this last case nothing will be plotted).
        support : str:
            If ``full`` use the finite end-points to set the limits of the plot. For unbounded
            end-points or if ``restricted`` use the 0.001 and 0.999 quantiles to set the limits.
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"legend"`` a
            ``"title"`` or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axes
        """
        if self.is_frozen:
            return plot_pdfpmf(
                self, moments, pointinterval, quantiles, support, legend, figsize, ax
            )
        else:
            raise ValueError(
                "Undefined distribution, "
                "you need to first define its parameters or use one of the fit methods"
            )

    def plot_cdf(
        self,
        moments=None,
        pointinterval=False,
        quantiles=None,
        support="full",
        legend="legend",
        figsize=None,
        ax=None,
    ):
        """
        Plot the cumulative distribution function.

        Parameters
        ----------
        moments : str
            Compute moments. Use any combination of the strings ``m``, ``d``, ``v``, ``s`` or ``k``
            for the mean (μ), standard deviation (σ), variance (σ²), skew (γ) or kurtosis (κ)
            respectively. Other strings will be ignored. Defaults to None.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        quantiles : list
            Values of the five quantiles to use when ``pointinterval=True`` if None (default)
            the values ``[0.05, 0.25, 0.5, 0.75, 0.95]`` will be used. The number of elements
            should be 5, 3, 1 or 0 (in this last case nothing will be plotted).
        support : str:
            If ``full`` use the finite end-points to set the limits of the plot. For unbounded
            end-points or if ``restricted`` use the 0.001 and 0.999 quantiles to set the limits.
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"legend"`` a
            ``"title"`` or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axes
        """
        if self.is_frozen:
            return plot_cdf(self, moments, pointinterval, quantiles, support, legend, figsize, ax)
        else:
            raise ValueError(
                "Undefined distribution, "
                "you need to first define its parameters or use one of the fit methods"
            )

    def plot_ppf(
        self,
        moments=None,
        pointinterval=False,
        quantiles=None,
        legend="legend",
        figsize=None,
        ax=None,
    ):
        """
        Plot the quantile function.

        Parameters
        ----------
        moments : str
            Compute moments. Use any combination of the strings ``m``, ``d``, ``v``, ``s`` or ``k``
            for the mean (μ), standard deviation (σ), variance (σ²), skew (γ) or kurtosis (κ)
            respectively. Other strings will be ignored. Defaults to None.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        quantiles : list
            Values of the five quantiles to use when ``pointinterval=True`` if None (default)
            the values ``[0.05, 0.25, 0.5, 0.75, 0.95]`` will be used. The number of elements
            should be 5, 3, 1 or 0 (in this last case nothing will be plotted).
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"legend"`` a
            ``"title"`` or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axes
        """
        if self.is_frozen:
            return plot_ppf(self, moments, pointinterval, quantiles, legend, figsize, ax)
        else:
            raise ValueError(
                "Undefined distribution, "
                "you need to first define its parameters or use one of the fit methods"
            )


class Continuous(Distribution):
    """Base class for continuous distributions."""

    def __init__(self):
        super().__init__()
        self.kind = "continuous"

    def xvals(self, support):
        """Provide x values in the support of the distribution. This is useful for example when
        plotting.

        Parameters
        ----------
        support : str
            Available options are "full" or "restricted".
        """
        lower_ep, upper_ep = self._finite_endpoints(support)
        x_vals = np.linspace(lower_ep, upper_ep, 1000)
        return x_vals

    def _fit_mle(self, sample, **kwargs):
        """
        Estimate the parameters of the distribution from a sample by maximizing the likelihood.

        Parameters
        ----------
        sample : array-like
            a sample
        kwargs : dict
            keywords arguments passed to scipy.stats.rv_continuous.fit
        """
        raise NotImplementedError


class Discrete(Distribution):
    """Base class for discrete distributions."""

    def __init__(self):
        super().__init__()
        self.kind = "discrete"

    def xvals(self, support):
        """Provide x values in the support of the distribution. This is useful for example when
        plotting.
        """
        lower_ep, upper_ep = self._finite_endpoints(support)
        x_vals = np.arange(lower_ep, upper_ep + 1, dtype=int)
        return x_vals



import logging
import numpy as np

from .utils.optimization import fit_to_sample

_log = logging.getLogger("preliz")


def mle(
    distributions,
    sample,
    plot=1,
    plot_kwargs=None,
    ax=None,
):
    """
    Find the maximum likelihood distribution with given a list of distributions
    and one sample.

    Parameters
    ----------
    distributions : list of PreliZ distribution
        Instance of a PreliZ distribution. Notice that the distributions will be
        updated inplace.
    sample : list or 1D array-like
        Data used to estimate the distribution parameters.
    plot : int
        Number of distributions to plots. Defaults to ``1`` (i.e. plot the best match)
        If larger than the number of passed distributions it will plot all of them.
        Use ``0`` or ``False`` to disable plotting.
    plot_kwargs : dict
        Dictionary passed to the method ``plot_pdf()`` of ``distribution``.
    ax : matplotlib axes

    Returns
    -------
    axes: matplotlib axes
    """
    sample = np.array(sample)
    x_min = sample.min()
    x_max = sample.max()

    fitted = fit_to_sample(distributions, sample, x_min, x_max)

    if plot > len(distributions):
        plot = len(distributions)

    if plot:
        idx = np.argsort(fitted.losses)
        idx = idx[:plot]
        for dist in fitted.distributions[idx]:
            if dist is not None:
                ax = dist.plot_pdf(plot_kwargs)
    return ax



"""
PreliZ.

Tools to help you pick a prior
"""
import logging

from matplotlib import rcParams

from .maxent import maxent
from .ppa import ppa
from .roulette import roulette
from .quartile import quartile
from .mle import mle
from .distributions import *

__all__ = ["maxent", "mle", "ppa", "roulette", "quartile"]

__version__ = "0.0.3"

_log = logging.getLogger("preliz")

if not logging.root.handlers:
    _log.setLevel(logging.INFO)
    if len(_log.handlers) == 0:
        handler = logging.StreamHandler()
        _log.addHandler(handler)

# Allow legend outside plot in maxent to be included when saving a figure
# We may want to make this more explicit by having preliz.rcParams
rcParams["savefig.bbox"] = "tight"



"""
Optimization routines and utilities
"""
from sys import modules
import numpy as np
from scipy.optimize import minimize, least_squares


def optimize_max_ent(dist, lower, upper, mass):
    def prob_bound(params, dist, lower, upper, mass):
        dist._update(*params)
        rv_frozen = dist.rv_frozen
        if dist.kind == "discrete":
            lower -= 1
        cdf0 = rv_frozen.cdf(lower)
        cdf1 = rv_frozen.cdf(upper)
        loss = (cdf1 - cdf0) - mass
        return loss

    def entropy_loss(params, dist):
        dist._update(*params)
        return -dist.rv_frozen.entropy()

    cons = {
        "type": "eq",
        "fun": prob_bound,
        "args": (dist, lower, upper, mass),
    }
    init_vals = dist.params
    bounds = dist.params_support
    if dist.name in ["halfstudent", "student"]:
        init_vals = init_vals[1:]
        bounds = bounds[1:]
    if dist.name == "skewnormal":
        init_vals = init_vals[:-1]
        bounds = bounds[:-1]
    if dist.name in ["betascaled", "truncatednormal"]:
        init_vals = init_vals[:-2]
        bounds = bounds[:-2]

    opt = minimize(entropy_loss, x0=init_vals, bounds=bounds, args=(dist), constraints=cons)
    dist._update(*opt["x"])

    return opt


def optimize_quartile(dist, x_vals):
    def func(params, dist, x_vals):
        dist._update(*params)
        loss = dist.rv_frozen.cdf(x_vals) - [0.25, 0.5, 0.75]
        return loss

    init_vals = dist.params
    if dist.name == "student":
        init_vals = init_vals[1:]
    if dist.name == "skewnormal":
        init_vals = init_vals[:-1]

    opt = least_squares(func, x0=init_vals, args=(dist, x_vals))
    dist._update(*opt["x"])

    return opt


def optimize_cdf(dist, x_vals, ecdf, **kwargs):
    def func(params, dist, x_vals, ecdf, **kwargs):
        dist._update(*params, **kwargs)
        loss = dist.rv_frozen.cdf(x_vals) - ecdf
        return loss

    init_vals = dist.params[:2]
    opt = least_squares(func, x0=init_vals, args=(dist, x_vals, ecdf), kwargs=kwargs)
    dist._update(*opt["x"])
    loss = opt["cost"]
    return loss


def optimize_matching_moments(dist, mean, sigma):
    def func(params, dist, mean, sigma):
        dist._update(*params)
        loss = ((dist.rv_frozen.mean() - mean) / mean) ** 2 + (
            (dist.rv_frozen.std() - sigma) / sigma
        ) ** 2
        return loss

    init_vals = dist.params
    opt = least_squares(func, x0=init_vals, args=(dist, mean, sigma))
    dist._update(*opt["x"])
    loss = opt["cost"]
    return loss


def optimize_ml(dist, sample):
    def negll(params, dist, sample):
        dist._update(*params)
        return -dist.rv_frozen.logpdf(sample).sum()

    dist._fit_moments(0, np.std(sample))
    init_vals = dist.params[::-1]

    opt = minimize(negll, x0=init_vals, bounds=dist.params_support, args=(dist, sample))

    dist._update(*opt["x"])

    return opt


def relative_error(dist, lower, upper, required_mass):
    if dist.kind == "discrete":
        lower -= 1
    computed_mass = dist.rv_frozen.cdf(upper) - dist.rv_frozen.cdf(lower)
    return abs((computed_mass - required_mass) / required_mass * 100), computed_mass


def get_distributions(dist_names):
    """
    Generate a subset of distributions which names agree with those in dist_names
    """
    dists = []
    for name in dist_names:
        dist = getattr(modules["preliz"], name)
        dists.append(dist())

    return dists


def fit_to_ecdf(selected_distributions, x_vals, ecdf, mean, std, x_min, x_max):
    """
    Minimize the difference between the cdf and the ecdf over a grid of values
    defined by x_min and x_max
    """
    fitted = Loss(len(selected_distributions))
    for dist in selected_distributions:
        kwargs = {}
        if dist.name == "betascaled":
            update_bounds_beta_scaled(dist, x_min, x_max)
            kwargs = {"lower": x_min, "upper": x_max}

        if dist._check_endpoints(x_min, x_max, raise_error=False):
            dist._fit_moments(mean, std)  # pylint:disable=protected-access
            loss = optimize_cdf(dist, x_vals, ecdf, **kwargs)

            fitted.update(loss, dist)

    return fitted.dist


def fit_to_sample(selected_distributions, sample, x_min, x_max):
    """
    Maximize the likelihood given a sample
    """
    fitted = Loss(len(selected_distributions))
    for dist in selected_distributions:
        if dist.name in ["betascaled", "truncatednormal"]:
            update_bounds_beta_scaled(dist, x_min, x_max)

        if dist._check_endpoints(x_min, x_max, raise_error=False):
            dist._fit_mle(sample)  # pylint:disable=protected-access
            if dist.kind == "continuous":
                loss = -dist.rv_frozen.logpdf(sample).sum()
            else:
                loss = -dist.rv_frozen.logpmf(sample).sum()

            fitted.update(loss, dist)

    return fitted


def update_bounds_beta_scaled(dist, x_min, x_max):
    dist.lower = x_min
    dist.upper = x_max
    dist.support = (x_min, x_max)
    return dist


class Loss:
    def __init__(self, size=None):
        self.old_loss = np.inf
        self.dist = None
        self.count = 0
        self.distributions = np.empty(size, dtype=object)
        self.losses = np.empty(size)

    def update(self, loss, dist):
        self.distributions[self.count] = dist
        self.losses[self.count] = loss
        self.count += 1
        if loss < self.old_loss:
            self.old_loss = loss
            self.dist = dist



"""
Discrete probability distributions.
"""
import logging
from math import ceil, floor

import numpy as np
from scipy import stats


from .distributions import Discrete
from ..utils.optimization import optimize_matching_moments

_log = logging.getLogger("preliz")

eps = np.finfo(float).eps


class Binomial(Discrete):
    R"""
    Binomial distribution.

    The discrete probability distribution of the number of successes
    in a sequence of n independent yes/no experiments, each of which
    yields success with probability p.

    The pmf of this distribution is

    .. math:: f(x \mid n, p) = \binom{n}{x} p^x (1-p)^{n-x}

    .. plot::
        :context: close-figs

        import arviz as az
        from preliz import Binomial
        az.style.use('arviz-white')
        ns = [5, 10, 10]
        ps = [0.5, 0.5, 0.7]
        for n, p in zip(ns, ps):
            Binomial(n, p).plot_pdf()

    ========  ==========================================
    Support   :math:`x \in \{0, 1, \ldots, n\}`
    Mean      :math:`n p`
    Variance  :math:`n p (1 - p)`
    ========  ==========================================

    Parameters
    ----------
    n : int
        Number of Bernoulli trials (n >= 0).
    p : float
        Probability of success in each trial (0 < p < 1).
    """

    def __init__(self, n=None, p=None):
        super().__init__()
        self.n = n
        self.p = p
        self.name = "binomial"
        self.params = (self.n, self.p)
        self.param_names = ("n", "p")
        self.params_support = ((eps, np.inf), (eps, 1 - eps))
        self.dist = stats.binom
        self.support = (0, np.inf)
        self._update_rv_frozen()

    def _get_frozen(self):
        frozen = None
        if any(self.params):
            frozen = self.dist(self.n, self.p)
        return frozen

    def _update(self, n, p):
        self.n = int(n)
        self.p = p
        self.params = (self.n, self.p)
        self.support = (0, self.n)
        self._update_rv_frozen()

    def _fit_moments(self, mean, sigma):
        # crude approximation for n and p
        n = mean + sigma * 2
        p = mean / n
        self._update(n, p)
        optimize_matching_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        # see https://doi.org/10.1016/j.jspi.2004.02.019 for details
        x_bar = np.mean(sample)
        x_std = np.std(sample)
        x_max = np.max(sample)
        n = ceil(x_max ** (1.5) * x_std / (x_bar**0.5 * (x_max - x_bar) ** 0.5))
        p = x_bar / n
        self._update(n, p)


class DiscreteUniform(Discrete):
    R"""
    Discrete Uniform distribution.

    The pmf of this distribution is

    .. math:: f(x \mid lower, upper) = \frac{1}{upper-lower+1}

    .. plot::
        :context: close-figs

        import arviz as az
        from preliz import DiscreteUniform
        az.style.use('arviz-white')
        ls = [1, -2]
        us = [6, 2]
        for l, u in zip(ls, us):
            ax = DiscreteUniform(l, u).plot_pdf()
            ax.set_ylim(0, 0.25)

    ========  ===============================================
    Support   :math:`x \in {lower, lower + 1, \ldots, upper}`
    Mean      :math:`\dfrac{lower + upper}{2}`
    Variance  :math:`\dfrac{(upper - lower + 1)^2 - 1}{12}`
    ========  ===============================================

    Parameters
    ----------
    lower: int
        Lower limit.
    upper: int
        Upper limit (upper > lower).
    """

    def __init__(self, lower=-np.inf, upper=np.inf):
        super().__init__()
        self.lower = lower
        self.upper = upper
        self.name = "discreteuniform"
        self.params = (self.lower, self.upper)
        self.param_names = ("lower", "upper")
        self.params_support = ((-np.inf, np.inf), (-np.inf, np.inf))
        self.dist = stats.randint
        self.support = (lower, upper)
        self.dist.a = -np.inf
        self.dist.b = np.inf
        self._update_rv_frozen()

    def _get_frozen(self):
        frozen = None
        if any(self.params):
            frozen = self.dist(self.lower, self.upper + 1)
        return frozen

    def _update(self, lower, upper):
        self.lower = floor(lower)
        self.upper = ceil(upper)
        self.params = (self.lower, self.upper)
        self.support = (self.lower, self.upper)
        self._update_rv_frozen()

    def _fit_moments(self, mean, sigma):
        spr = (12 * sigma**2 + 1) ** 0.5
        lower = 0.5 * (2 * mean - spr + 1)
        upper = 0.5 * (2 * mean + spr - 1)
        self._update(lower, upper)

    def _fit_mle(self, sample):
        lower = np.min(sample)
        upper = np.max(sample)
        self._update(lower, upper)


class NegativeBinomial(Discrete):
    R"""
    Negative binomial distribution.

    If it is parametrized in terms of n and p, the negative binomial describes
    the probability to have x failures before the n-th success, given the
    probability p of success in each trial. Its pmf is

    .. math::

        f(x \mid n, p) =
           \binom{x + n - 1}{x}
           (p)^n (1 - p)^x

    .. plot::
        :context: close-figs

        import arviz as az
        from preliz import NegativeBinomial
        az.style.use('arviz-white')
        ns = [5, 10, 10]
        ps = [0.5, 0.5, 0.7]
        for n, p in zip(ns, ps):
            NegativeBinomial(n, p).plot_pdf()

    ========  ==========================
    Support   :math:`x \in \mathbb{N}_0`
    Mean      :math:`\frac{(1-p) r}{p}`
    Variance  :math:`\frac{(1-p) r}{p^2}`
    ========  ==========================

    Parameters
    ----------
    n: float
        Number of target success trials (n > 0)
    p: float
        Probability of success in each trial (0 < p < 1).
    """

    def __init__(self, n=None, p=None):
        super().__init__()
        self.n = n
        self.p = p
        self.name = "negativebinomial"
        self.params = (self.n, self.p)
        self.param_names = ("n", "p")
        self.params_support = ((eps, np.inf), (eps, 1 - eps))
        self.dist = stats.nbinom
        self.support = (0, np.inf)
        self._update_rv_frozen()

    def _get_frozen(self):
        frozen = None
        if any(self.params):
            frozen = self.dist(self.n, self.p)
        return frozen

    def _update(self, n, p):
        self.n = n
        self.p = p
        self.params = (self.n, p)
        self._update_rv_frozen()

    def _fit_moments(self, mean, sigma):
        n = mean**2 / (sigma**2 - mean)
        p = mean / sigma**2
        self._update(n, p)

    def _fit_mle(self, sample):
        # the upper bound is based on a quick heuristic. The fit will underestimate
        # the value of n when p is very close to 1.
        fitted = stats.fit(self.dist, sample, bounds={"n": (1, max(sample) * 5)})
        if not fitted.success:
            _log.info("Optimization did not terminate successfully.")
        self._update(fitted.params.n, fitted.params.p)  # pylint: disable=no-member


class Poisson(Discrete):
    R"""
    Poisson log-likelihood.

    Often used to model the number of events occurring in a fixed period
    of time when the times at which events occur are independent.
    The pmf of this distribution is

    .. math:: f(x \mid \mu) = \frac{e^{-\mu}\mu^x}{x!}

    .. plot::
        :context: close-figs

        import arviz as az
        from preliz import Poisson
        az.style.use('arviz-white')
        for mu in [0.5, 3, 8]:
            Poisson(mu).plot_pdf()

    ========  ==========================
    Support   :math:`x \in \mathbb{N}_0`
    Mean      :math:`\mu`
    Variance  :math:`\mu`
    ========  ==========================

    Parameters
    ----------
    mu: float
        Expected number of occurrences during the given interval
        (mu >= 0).

    Notes
    -----
    The Poisson distribution can be derived as a limiting case of the
    binomial distribution.
    """

    def __init__(self, mu=None):
        super().__init__()
        self.mu = mu
        self.name = "poisson"
        self.params = (self.mu,)
        self.param_names = ("mu",)
        self.params_support = ((eps, np.inf),)
        self.dist = stats.poisson
        self.support = (0, np.inf)
        self._update_rv_frozen()

    def _get_frozen(self):
        frozen = None
        if any(self.params):
            frozen = self.dist(self.mu)
        return frozen

    def _update(self, mu):
        self.mu = mu
        self.params = (self.mu,)
        self._update_rv_frozen()

    def _fit_moments(self, mean, sigma=None):  # pylint: disable=unused-argument
        self._update(mean)

    def _fit_mle(self, sample):
        mu = np.mean(sample)
        self._update(mu)



import pytest
from numpy.testing import assert_allclose

import preliz as pz


@pytest.mark.parametrize(
    "distribution, params",
    [
        (pz.Beta, (2, 5)),
        (pz.BetaScaled, (2, 5, -1, 4)),
        (pz.Cauchy, (0, 1)),
        (pz.ChiSquared, (1,)),
        (pz.Exponential, (5,)),
        (pz.Gamma, (2, 5)),
        (pz.HalfCauchy, (1,)),
        (pz.HalfNormal, (1,)),
        (pz.HalfStudent, (3, 1)),
        (pz.HalfStudent, (1000000, 1)),
        (pz.HalfNormal, (2,)),
        (pz.InverseGamma, (3, 5)),
        (pz.Laplace, (0, 2)),
        (pz.LogNormal, (0, 1)),
        (pz.Normal, (0, 1)),
        (pz.Pareto, (5, 1)),
        (pz.SkewNormal, (0, 1, 6)),
        (pz.Student, (4, 0, 1)),
        (pz.TruncatedNormal, (0, 1, -1, 1)),
        (pz.Uniform, (2, 5)),
        (pz.VonMises, (1, 2)),
        (pz.Wald, (2, 1)),
        (pz.Weibull, (2, 1)),
        (pz.Binomial, (5, 0.5)),
        (pz.DiscreteUniform, (-2, 2)),
        (pz.NegativeBinomial, (10, 0.5)),
        (pz.Poisson, (4.2,)),
    ],
)
def test_auto_recover(distribution, params):
    dist = distribution(*params)
    sample = dist.rvs(10000)
    pz.mle([dist], sample)
    assert_allclose(dist.params, params, atol=1)


def test_recover_right():
    dists = [pz.Normal(), pz.Gamma(), pz.Poisson()]
    sample = pz.Normal(0, 1).rvs(10000)
    pz.mle(dists, sample)
    assert dists[0].name == "normal"

    sample = pz.Gamma(2, 10).rvs(10000)
    pz.mle(dists, sample)
    assert dists[1].name == "gamma"

    sample = pz.Poisson(10).rvs(10000)
    pz.mle(dists, sample)
    assert dists[2].name == "poisson"



from .continuous import *
from .discrete import *

all_continuous = [
    Beta,
    BetaScaled,
    Cauchy,
    ChiSquared,
    Exponential,
    Gamma,
    HalfCauchy,
    HalfNormal,
    HalfStudent,
    InverseGamma,
    Laplace,
    LogNormal,
    Normal,
    Pareto,
    SkewNormal,
    Student,
    TruncatedNormal,
    Uniform,
    VonMises,
    Wald,
    Weibull,
]
all_discrete = [Binomial, DiscreteUniform, NegativeBinomial, Poisson]


__all__ = [s.__name__ for s in all_continuous] + [s.__name__ for s in all_discrete]



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import _pylab_helpers
from scipy.stats._distn_infrastructure import rv_continuous_frozen, rv_discrete_frozen
from scipy.interpolate import interp1d, PchipInterpolator


def plot_pointinterval(distribution, quantiles=None, rotated=False, ax=None):
    """
    By default plot the quantiles [0.05, 0.25, 0.5, 0.75, 0.95]
    The median as dot and the two interquantiles ranges (0.05-0.95) and (0.25-0.75) as lines.

    Parameters
    ----------
    distribution : preliz distribution or array
    quantiles : list
        The number of elements should be 5, 3, 1 or 0 (in this last case nothing will be plotted).
        defaults to [0.05, 0.25, 0.5, 0.75, 0.95].
    rotated : bool
        Whether to do the plot along the x-axis (default) or on the y-axis
    ax : matplotlib axis
    """
    if quantiles is None:
        quantiles = [0.05, 0.25, 0.5, 0.75, 0.95]

    if isinstance(distribution, (rv_continuous_frozen, rv_discrete_frozen)):
        q_s = distribution.ppf(quantiles).tolist()
    else:
        q_s = np.quantile(distribution, quantiles).tolist()

    q_s_size = len(q_s)
    if not q_s_size in (5, 3, 1, 0):
        raise ValueError("quantiles should have 5, 3, 1 or 0 elements")

    if rotated:
        if q_s_size == 5:
            ax.plot([0, 0], (q_s.pop(0), q_s.pop(-1)), "k", solid_capstyle="butt", lw=1.5)
        if q_s_size > 2:
            ax.plot([0, 0], (q_s.pop(0), q_s.pop(-1)), "k", solid_capstyle="butt", lw=4)
        if q_s_size > 0:
            ax.plot(0, q_s[0], "wo", mec="k")
    else:
        if q_s_size == 5:
            ax.plot((q_s.pop(0), q_s.pop(-1)), [0, 0], "k", solid_capstyle="butt", lw=1.5)
        if q_s_size > 2:
            ax.plot((q_s.pop(0), q_s.pop(-1)), [0, 0], "k", solid_capstyle="butt", lw=4)

        if q_s_size > 0:
            ax.plot(q_s[0], 0, "wo", mec="k")


def plot_pdfpmf(dist, moments, pointinterval, quantiles, support, legend, figsize, ax):
    ax = get_ax(ax, figsize)
    color = next(ax._get_lines.prop_cycler)["color"]
    label = repr_to_matplotlib(dist)
    if moments is not None:
        label += get_moments(dist, moments)

    x = dist.xvals(support)
    if dist.kind == "continuous":
        density = dist.rv_frozen.pdf(x)
        ax.plot(x, density, label=label, color=color)
        ax.get_ylim()
        ax.set_yticks([])
    else:
        mass = dist.rv_frozen.pmf(x)
        eps = np.clip(dist._finite_endpoints(support), *dist.support)
        x_c = np.linspace(*eps, 1000)

        if len(x) > 2:
            interp = PchipInterpolator(x, mass)
        else:
            interp = interp1d(x, mass)

        mass_c = np.clip(interp(x_c), np.min(mass), np.max(mass))

        ax.plot(x_c, mass_c, ls="dotted", color=color)
        ax.plot(x, mass, "o", label=label, color=color)

    if pointinterval:
        plot_pointinterval(dist.rv_frozen, quantiles=quantiles, ax=ax)

    if legend == "title":
        ax.set_title(label)
    elif legend == "legend":
        side_legend(ax)
    return ax


def plot_cdf(dist, moments, pointinterval, quantiles, support, legend, figsize, ax):
    ax = get_ax(ax, figsize)
    color = next(ax._get_lines.prop_cycler)["color"]
    label = repr_to_matplotlib(dist)
    if moments is not None:
        label += get_moments(dist, moments)

    eps = dist._finite_endpoints(support)
    x = np.linspace(*eps, 1000)
    cdf = dist.rv_frozen.cdf(x)
    ax.plot(x, cdf, label=label, color=color)

    if pointinterval:
        plot_pointinterval(dist.rv_frozen, quantiles=quantiles, ax=ax)

    if legend == "title":
        ax.set_title(label)
    elif legend == "legend":
        side_legend(ax)
    return ax


def plot_ppf(dist, moments, pointinterval, quantiles, legend, figsize, ax):
    ax = get_ax(ax, figsize)
    color = next(ax._get_lines.prop_cycler)["color"]
    label = repr_to_matplotlib(dist)
    if moments is not None:
        label += get_moments(dist, moments)

    x = np.linspace(0, 1, 1000)
    ax.plot(x, dist.rv_frozen.ppf(x), label=label, color=color)

    if pointinterval:
        plot_pointinterval(dist.rv_frozen, quantiles=quantiles, rotated=True, ax=ax)

    if legend == "title":
        ax.set_title(label)
    elif legend == "legend":
        side_legend(ax)
    return ax


def get_ax(ax, figsize):
    if ax is None:
        fig_manager = _pylab_helpers.Gcf.get_active()
        if fig_manager is not None:
            ax = fig_manager.canvas.figure.gca()
        else:
            _, ax = plt.subplots(figsize=figsize)
    return ax


def side_legend(ax):
    bbox = ax.get_position()
    ax.set_position([bbox.x0, bbox.y0, bbox.width, bbox.height])
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))


def repr_to_matplotlib(distribution):
    string = distribution.__repr__()
    string = string.replace("\x1b[1m", r"$\bf{")
    string = string.replace("\x1b[0m", "}$")
    return string


def get_moments(dist, moments):
    names = {
        "m": "μ",
        "d": "σ",
        "s": "γ",
        "v": "σ²",
        "k": "κ",
    }
    str_m = []
    seen = []
    for moment in moments:
        if moment not in seen:
            if moment == "d":
                value = dist.rv_frozen.stats("v") ** 0.5
            else:
                value = dist.rv_frozen.stats(moment)
            if isinstance(value, (np.ndarray, int, float)):
                str_m.append(f"{names[moment]}={value:.3g}")

        seen.append(moment)

    return "\n" + ", ".join(str_m)



"""Prior predictive check assistant."""

import logging

import arviz as az
import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial import KDTree


from preliz import distributions
from .utils.plot_utils import plot_pointinterval, repr_to_matplotlib

_log = logging.getLogger("preliz")

pymc_to_preliz = {
    "normal": "Normal",
    "halfnormal": "HalfNormal",
}


def ppa(idata, model, summary="octiles"):
    """
    Prior predictive check assistant.

    This is experimental

    Parameters
    ----------

    idata : InferenceData
        With at least the `prior` and `prior_predictive` groups
    model : PyMC model
        Model associated to ``idata``.
    summary : str:
        Summary statistics applied to prior samples in order to define (dis)similarity
        of distributions. Current options are `octiles`, `hexiles`, `quantiles`,
        `sort` (sort data) `octi_sum` (robust estimation of first 4 moments from octiles).
    """
    _log.info(
        """Enter at your own risk."""
        """This is highly experimental code and not recommended for regular use."""
    )

    try:
        shell = get_ipython().__class__.__name__  # pylint:disable=undefined-variable
        if shell == "ZMQInteractiveShell" and "nbagg" not in get_backend():
            _log.info(
                "To run roulette you need Jupyter notebook, or Jupyter lab."
                "You will also need to use the magic `%matplotlib widget`"
            )
    except NameError:
        pass

    shown = []
    obs_rv = model.observed_RVs[0].name  # only one observed variable for the moment
    pp_samples = idata.prior_predictive[obs_rv].squeeze().values
    prior_samples = idata.prior.squeeze()
    sample_size = pp_samples.shape[0]
    pp_summary, kdt = compute_summaries(pp_samples, summary)
    pp_samples_idxs, shown = initialize_subsamples(pp_summary, shown, kdt)
    fig, axes = plot_samples(pp_samples, pp_samples_idxs)

    clicked = []
    selected = []
    choices = []

    output = widgets.Output()

    with output:
        button_carry_on = widgets.Button(description="carry on")
        button_return_prior = widgets.Button(description="return prior")

        def carry_on_(_):
            carry_on(fig, axes, clicked, pp_samples, pp_summary, choices, selected, shown, kdt)

        button_carry_on.on_click(carry_on_)

        def on_return_prior_(_):
            on_return_prior(fig, selected, model, sample_size)

        button_return_prior.on_click(on_return_prior_)

        def click(event):
            if event.inaxes not in clicked:
                clicked.append(event.inaxes)
            else:
                clicked.remove(event.inaxes)
                plt.setp(event.inaxes.spines.values(), color="k", lw=1)

            for ax in clicked:
                plt.setp(ax.spines.values(), color="C1", lw=3)
            fig.canvas.draw()

        def on_return_prior(fig, selected, model, sample_size):
            selected = np.unique(selected)
            num_selected = len(selected)

            if num_selected > 1:
                string = (
                    f"You have selected {num_selected} out of {sample_size} prior"
                    "predictive samples\nThat selection is consistent with the"
                    "following priors:\n"
                )

                subsample = select_prior_samples(selected, prior_samples, model)
                string = back_fitting(model, subsample, string)

                fig.clf()
                plt.text(0.2, 0.5, string, fontsize=14)
                plt.yticks([])
                plt.xticks([])
            else:
                fig.suptitle("Please select more distributions", fontsize=16)

            fig.canvas.draw()

        fig.canvas.mpl_connect("button_press_event", click)

    controls = widgets.VBox([button_carry_on, button_return_prior])

    display(widgets.HBox([controls]))  # pylint:disable=undefined-variable


def carry_on(fig, axes, clicked, pp_samples, pp_summary, choices, selected, shown, kdt):
    choices.extend([int(ax.get_title()) for ax in clicked])
    selected.extend(choices)

    fig.suptitle("")
    for ax in clicked:
        plt.setp(ax.spines.values(), color="k", lw=1)
    for ax in axes:
        ax.cla()
    for ax in list(clicked):
        clicked.remove(ax)

    pp_samples_idxs, shown = keep_sampling(pp_summary, choices, shown, kdt)
    if not pp_samples_idxs:
        pp_samples_idxs, shown = initialize_subsamples(pp_summary, shown, kdt)
    fig, _ = plot_samples(pp_samples, pp_samples_idxs, fig)
    fig.canvas.draw()


def compute_summaries(pp_samples, summary):
    if summary == "octiles":
        pp_summary = np.quantile(
            pp_samples, [0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875], axis=1
        ).T
    elif summary == "quartiles":
        pp_summary = np.quantile(
            pp_samples,
            [
                0.25,
                0.5,
                0.75,
            ],
            axis=1,
        ).T

    elif summary == "hexiles":
        pp_summary = np.quantile(pp_samples, [0.25, 0.375, 0.5, 0.625, 0.75], axis=1).T
    elif summary == "sort":
        pp_summary = np.sort(pp_samples)
    elif summary == "octi_sum":
        suma = np.quantile(pp_samples, [0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875], axis=1).T
        s_a = suma[:, 3]
        s_b = suma[:, 5] - suma[:, 1]
        s_g = (suma[:, 5] + suma[:, 1] - 2 * suma[:, 3]) / s_b
        s_k = (suma[:, 6] - suma[:, 4] + suma[:, 2] - suma[:, 0]) / s_b
        pp_summary = np.stack([s_a, s_b, s_g, s_k]).T

    kdt = KDTree(pp_summary)
    return pp_summary, kdt


def initialize_subsamples(pp_summary, shown, kdt):
    new = np.random.choice(list(set(range(0, len(pp_summary))) - set(shown)))
    samples = [new]

    for _ in range(8):
        farthest_neighbor = pp_summary.shape[0]
        while new in samples:
            _, new = kdt.query(pp_summary[samples[-1]], [farthest_neighbor])
            new = new.item()
            farthest_neighbor -= 1
        samples.append(new)

    shown.extend(samples)

    return samples, shown


def keep_sampling(pp_summary, choices, shown, kdt):
    """
    Find distribution similar to the ones in `choices`, but not already shown.
    If `choices` is empty return an empty selection.
    """
    if choices:
        new = choices.pop(0)
        samples = [new]

        for _ in range(9):
            nearest_neighbor = 2
            while new in samples or new in shown:
                _, new = kdt.query(pp_summary[samples[-1]], [nearest_neighbor])
                new = new.item()
                nearest_neighbor += 1
            samples.append(new)

        shown.extend(samples[1:])

        return samples[1:], shown
    else:
        return [], shown


def plot_samples(pp_samples, pp_samples_idxs, fig=None):
    row_colum = int(np.ceil(len(pp_samples_idxs) ** 0.5))

    if fig is None:
        fig, axes = plt.subplots(row_colum, row_colum, figsize=(8, 6), sharex=True)
    else:
        axes = np.array(fig.axes)

    try:
        axes = axes.ravel()
    except AttributeError:
        axes = [axes]

    for ax, idx in zip(axes, pp_samples_idxs):
        ax.axvline(0, ls="--", color="0.5")

        sample = pp_samples[idx]
        az.plot_kde(sample, ax=ax, plot_kwargs={"color": "C0"})

        plot_pointinterval(sample, ax=ax)
        ax.set_title(idx)
        ax.set_yticks([])

    return fig, axes


def select_prior_samples(selected, prior_samples, model):
    """
    Given a selected set of prior predictive samples pick the corresponding
    prior samples.
    """
    rvs = model.unobserved_RVs  # we should exclude deterministics
    rv_names = [rv.name for rv in rvs]
    subsample = {rv: prior_samples[rv].sel(draw=selected).values.squeeze() for rv in rv_names}

    return subsample


def back_fitting(model, subset, string):
    """
    Use MLE to fit a subset of the prior samples to the individual prior distributions
    """
    for unobserved in model.unobserved_RVs:
        distribution = unobserved.owner.op.name
        dist = getattr(distributions, pymc_to_preliz[distribution])()
        name = unobserved.name
        idx = unobserved.owner.nin - 3
        params = unobserved.owner.inputs[-idx:]
        is_fitable = any(param.name is None for param in params)
        if is_fitable:
            dist._fit_mle(subset[name])
            string += f"{repr_to_matplotlib(dist)}\n"
    return string



import pytest
import numpy as np

from numpy.testing import assert_allclose

from preliz.distributions.continuous import (
    HalfCauchy,
    HalfNormal,
    HalfStudent,
    TruncatedNormal,
    InverseGamma,
    VonMises,
)
from preliz import maxent
from preliz.distributions import (
    Beta,
    Cauchy,
    ChiSquared,
    Exponential,
    Gamma,
    Laplace,
    LogNormal,
    Normal,
    Pareto,
    Student,
    Uniform,
    Wald,
    Weibull,
    DiscreteUniform,
    Poisson,
)


@pytest.mark.parametrize(
    "distribution, name, lower, upper, mass, nu, support, result",
    [
        (Beta, "beta", 0.2, 0.6, 0.9, None, (0, 1), (6.112, 9.101)),
        (Cauchy, "cauchy", -1, 1, 0.6, None, (-np.inf, np.inf), (0, 0.726)),
        (ChiSquared, "chisquared", 0, 4, 0.9, 1, (0, np.inf), (1.659)),
        (Exponential, "exponential", 0, 4, 0.9, None, (0, np.inf), (0.575)),
        (Gamma, "gamma", 0, 10, 0.7, None, (0, np.inf), (0.868, 0.103)),
        (HalfCauchy, "halfcauchy", 0, 10, 0.7, None, (0, np.inf), (5.095)),
        (HalfNormal, "halfnormal", 0, 10, 0.7, None, (0, np.inf), (9.648)),
        (HalfStudent, "halfstudent", 0, 10, 0.7, 3, (0, np.inf), (8.001)),
        (HalfStudent, "halfstudent", 0, 10, 0.7, 10000, (0, np.inf), (9.648)),
        (InverseGamma, "inversegamma", 0, 1, 0.99, None, (0, np.inf), (8.889, 3.439)),
        (Laplace, "laplace", -1, 1, 0.9, None, (-np.inf, np.inf), (0, 0.435)),
        (LogNormal, "lognormal", 1, 4, 0.5, None, (0, np.inf), (1.216, 0.859)),
        (Normal, "normal", -1, 1, 0.683, None, (-np.inf, np.inf), (0, 1)),
        (Normal, "normal", 10, 12, 0.99, None, (-np.inf, np.inf), (11, 0.388)),
        (Pareto, "pareto", 1, 4, 0.9, None, (1, np.inf), (1.660, 1)),
        (Student, "student", -1, 1, 0.683, 4, (-np.inf, np.inf), (0, 0.875)),
        (Student, "student", -1, 1, 0.683, 10000, (-np.inf, np.inf), (0, 1)),
        (TruncatedNormal, "truncatednormal", -1, 1, 0.683, None, (-np.inf, np.inf), (0, 1)),
        (Uniform, "uniform", -2, 10, 0.9, None, (-np.inf, np.inf), (-2.666, 10.666)),
        (VonMises, "vonmises", -1, 1, 0.9, None, (-np.inf, np.inf), (0.0, 3.294)),
        (Wald, "wald", 0, 10, 0.9, None, (0, np.inf), (5.061, 7.937)),
        (Weibull, "weibull", 0, 10, 0.9, None, (0, np.inf), (1.411, 5.537)),
        (DiscreteUniform, "discreteuniform", -2, 10, 0.9, None, (-3, 11), (-2, 10)),
        (Poisson, "poisson", 0, 3, 0.7, None, (0, np.inf), (2.763)),
    ],
)
def test_maxent(distribution, name, lower, upper, mass, nu, support, result):
    if nu is None:
        dist = distribution()
    else:
        dist = distribution(nu=nu)
    _, opt = maxent(dist, lower, upper, mass)
    rv_frozen = dist.rv_frozen

    assert rv_frozen.name == name

    if dist.name == "pareto":
        assert pytest.approx(rv_frozen.support(), 0.3) == support
    else:
        assert rv_frozen.support() == support

    if dist.name != "discreteuniform":  # optimization fails to converge, but results are reasonable
        assert opt.success
    assert_allclose(opt.x, result, atol=0.001)



import pytest
from numpy.testing import assert_almost_equal
import numpy as np

from preliz.distributions import (
    Beta,
    Cauchy,
    ChiSquared,
    Gamma,
    Exponential,
    HalfCauchy,
    HalfNormal,
    HalfStudent,
    InverseGamma,
    Laplace,
    LogNormal,
    Normal,
    Pareto,
    SkewNormal,
    Student,
    TruncatedNormal,
    Uniform,
    VonMises,
    Wald,
    Weibull,
    Binomial,
    DiscreteUniform,
    NegativeBinomial,
    Poisson,
)


@pytest.mark.parametrize(
    "distribution, params",
    [
        (Normal, (0, 1)),
        (Beta, (2, 5)),
        (ChiSquared, (1,)),
        (Gamma, (1, 0.5)),
        (HalfNormal, (1,)),
        (HalfStudent, (3, 1)),
        (HalfStudent, (1000000, 1)),
        (InverseGamma, (3, 1)),
        (Laplace, (0, 1)),
        (LogNormal, (0, 0.5)),
        (Exponential, (0.5,)),
        (Pareto, (5, 1)),
        (SkewNormal, (0, 1, 0)),
        (Student, (3, 0, 1)),
        (Student, (1000000, 0, 1)),
        (TruncatedNormal, (0, 1, -np.inf, np.inf)),
        (Uniform, (0, 1)),
        (VonMises, (0, 1000)),
        (Wald, (1, 1)),
        (Weibull, (2, 1)),
        (Binomial, (2, 0.5)),
        (Binomial, (2, 0.1)),
        (NegativeBinomial, (2, 0.7)),
        (NegativeBinomial, (2, 0.3)),
        (Poisson, (4.5,)),
        (DiscreteUniform, (0, 1)),
    ],
)
def test_moments(distribution, params):
    dist = distribution(*params)
    if "student" in dist.name:
        dist_ = distribution(nu=params[0])
    else:
        dist_ = distribution()

    dist_._fit_moments(dist.rv_frozen.mean(), dist.rv_frozen.std())

    tol = 5
    if dist.name == "binomial":
        tol = 0
    assert_almost_equal(dist.rv_frozen.mean(), dist_.rv_frozen.mean(), tol)
    assert_almost_equal(dist.rv_frozen.std(), dist_.rv_frozen.std(), tol)
    assert_almost_equal(params, dist_.params, 0)


@pytest.mark.parametrize(
    "distribution, params",
    [
        (Normal, (0, 1)),
        (Beta, (2, 5)),
        (Cauchy, (0, 1)),
        (ChiSquared, (1,)),
        (Gamma, (1, 0.5)),
        (HalfCauchy, (1,)),
        (HalfNormal, (1,)),
        (HalfStudent, (3, 1)),
        (InverseGamma, (3, 0.5)),
        (Laplace, (0, 1)),
        (LogNormal, (0, 0.5)),
        (Exponential, (0.5,)),
        (Pareto, (5, 1)),
        (SkewNormal, (0, 1, 0)),
        (SkewNormal, (0, 1, -1)),
        (Student, (4, 0, 1)),
        (Student, (1000, 0, 1)),
        (TruncatedNormal, (0, 1, -1, 1)),
        (Uniform, (0, 1)),
        (VonMises, (0, 1)),
        (Wald, (1, 1)),
        (Weibull, (2, 1)),
        (Binomial, (2, 0.5)),
        (Binomial, (2, 0.1)),
        (NegativeBinomial, (2, 0.7)),
        (NegativeBinomial, (2, 0.3)),
        (Poisson, (4.5,)),
        (DiscreteUniform, (0, 1)),
    ],
)
def test_mle(distribution, params):
    dist = distribution(*params)
    sample = dist.rv_frozen.rvs(20000)
    dist_ = distribution()
    dist_._fit_mle(sample)

    assert_almost_equal(dist.rv_frozen.mean(), dist_.rv_frozen.mean(), 1)
    assert_almost_equal(dist.rv_frozen.std(), dist_.rv_frozen.std(), 1)
    if dist.name == "student":
        assert_almost_equal(params[1:], dist_.params[1:], 0)
    else:
        assert_almost_equal(params, dist_.params, 0)


@pytest.mark.parametrize("fmt", (".2f", ".1g"))
@pytest.mark.parametrize("mass", (0.5, 0.95))
def test_summary(fmt, mass):
    result = Normal(0, 1).summary(fmt, mass)
    assert result.mean == 0
    assert result._fields == ("mean", "median", "std", "lower", "upper")
    result = Poisson(2).summary()
    assert result.mean == 2


@pytest.mark.parametrize(
    "distribution, params, alt_names",
    [
        (Beta, (2, 5), ("mu", "sigma")),
        (Beta, (5, 2), ("mu", "kappa")),
        (Gamma, (2, 1), ("mu", "sigma")),
        (InverseGamma, (0, 2), ("mu", "sigma")),
        (Normal, (0, 1), ("mu", "tau")),
        (Student, (1000, 0, 1), ("nu", "mu", "lam")),
    ],
)
def test_alternative_parametrization(distribution, params, alt_names):
    dist0 = distribution(*params)
    params1 = {p: getattr(dist0, p) for p in alt_names}
    dist1 = distribution(**params1)
    assert_almost_equal(dist0.params, dist1.params)
    for p in alt_names:
        assert p in dist1.__repr__()



import numpy as np
from scipy.special import gamma


def hdi_from_pdf(dist, mass=0.95):
    """
    Approximate the HDI by evaluating the pdf.
    This is faster, but potentially less accurate, than directly minimizing the
    interval as evaluating the ppf can be slow, specially for some distributions.
    """
    if dist.kind == "continuous":
        lower_ep, upper_ep = dist._finite_endpoints("full")
        x_vals = np.linspace(lower_ep, upper_ep, 10000)
        pdf = dist.rv_frozen.pdf(x_vals)
        pdf = pdf / pdf.sum()
    else:
        x_vals = dist.xvals(support="full")
        pdf = dist.rv_frozen.pmf(x_vals)

    sorted_idx = np.argsort(pdf)[::-1]
    mass_cum = 0
    indices = []
    for idx in sorted_idx:
        mass_cum += pdf[idx]
        indices.append(idx)
        if mass_cum >= mass:
            break
    return x_vals[np.sort(indices)[[0, -1]]]


def garcia_approximation(mean, sigma):
    """
    Approximate method of moments for Weibull distribution, provides good results for values of
    alpha larger than 0.83.

    Oscar Garcia. Simplified method-of-moments estimation for the Weibull distribution. 1981.
    New Zealand Journal of Forestry Science 11:304-306
    https://www.scionresearch.com/__data/assets/pdf_file/0010/36874/NZJFS1131981GARCIA304-306.pdf
    """
    ks = [-0.221016417, 0.010060668, 0.117358987, -0.050999126]  # pylint: disable=invalid-name
    z = sigma / mean  # pylint: disable=invalid-name

    poly = 0
    for idx, k in enumerate(ks):
        poly += k * z**idx

    alpha = 1 / (z * (1 + (1 - z) ** 2 * poly))
    beta = 1 / (gamma(1 + 1 / alpha) / (mean))
    return alpha, beta
